module datbyread
  use estim
  use paths
  use mpimod
  implicit none
  type pnrobs
     integer :: pnr !these maintained for future merge operations.
     integer :: female
     integer :: imm
     integer :: elen
     integer :: birthyr
     integer :: gradyr
     integer :: lagstartspell
     integer :: lagendspell
  end type pnrobs
  type cvrobs
     integer :: cvr
     integer :: sector
     integer :: year
     integer :: industry 
     integer :: bweek !firm's birth week
  end type cvrobs
  type spellobs
     integer :: pnr
     real(8) :: hours
     integer :: spellstart
     integer :: start
     integer :: end
     integer :: cvr
     real(8) :: earnings
     integer :: yr
  end type spellobs
  type hist
     type(spellobs) :: so
     type(hist), pointer :: prev=>null(), next=>null()
  end type hist
  type(hist), pointer :: first,last
contains
  recursive subroutine remove_hist(myhist) !Removes this and all elements following it.
    type(hist), pointer :: myhist, next
    next => myhist%next
    do while(associated(next))
       call remove_hist(next)
    end do
    deallocate(myhist)
  end subroutine remove_hist

  subroutine addhist(myspobs) !add a spell to this, and make this point to the newly added spell.
    type(spellobs), intent(in) :: myspobs
    type(hist), pointer :: new
    if(associated(first)) then
       allocate(new)
       new%so=myspobs
       new%prev=>last !create backward link
       last%next=>new !create forward link
       nullify(new%next)
       last=>new
    else
       allocate(first)
       first%so=myspobs
       last=>first
    end if
  end subroutine addhist

  subroutine readdat(name)
    character(5), intent(in) :: name
    type(spell), pointer :: sp
    type(spellobs) :: spobs
    character(20) :: mylabel(8),yrlabel(80)
    integer, parameter :: recalltol=20 !If worker returns to old employer within this time, join spells.
    integer, parameter :: minunemdur=3 !If worker returns to old employer within this time, join spells.
    integer :: info,ii,mypad,pad
    integer :: wc !worker count (and also index in work(:))    
    wc=1
    mypad=0

    open(10,file=datpath//"newyearweek_jan.csv",status='old')
    read(10,*) yrlabel
    read(10,*) info,(yrtoweek(ii),ii=1,79)
    close(10)
    yrtoweek(80)=yrtoweek(79)+52

    call readpnr !populates work and dim_I
    call readcvr !populates firm and dim_J

    open(10,file=datpath//name//"_spell.csv",status='old')
    read(10,*) mylabel
    read(10,*,iostat=info) spobs
    spobs%yr=spobs%yr-1934
    call readhist !read worker histories and commit to memory. Upon return end of file has been reached.
    close(10)
    call mpi_allreduce(mypad,pad,1,mpi_integer,mpi_sum,mpi_comm_world,ierr)
	  if(master) write(6,'("Number of terminal padded worker histories:",i12," out of total number of workers",i12)') pad, dim_I
  contains
    recursive subroutine readhist !use spobs as global variable.
      integer :: mypnr
      if(spobs%pnr==-1) return !we are done
      if(associated(first)) then !clean up, if necessary.
         call remove_hist(first)
         nullify(first,last)
      end if
      mypnr=spobs%pnr !record worker pnr

      do while(spobs%pnr==mypnr) !keep adding to linked list as long as spobs belongs to mypnr history.
         if(spobs%hours>0.d0) call addhist(spobs) !only add spells with positive hours worked.
         read(10,*,iostat=info) spobs
         spobs%yr=spobs%yr-1934 !make 1935 year 1.
         if(info<0) then !reached end of file
            spobs%pnr=-1
         end if
      end do

      do while(work(wc)%id<mypnr) !first find work(.) index to store history in.
         wc=wc+1
      end do
      if(work(wc)%id/=mypnr) stop'fatal(commit_hist_to_mem): failed to find worker id with matching pnr,'
      !			if(master) print *,wc

      if(mod(wc-1,numprocs)==myid) call commit_to_mem !take raw worker history and organize it in spell structure.
      call readhist !Do it again. At this point spobs contains a new pnr.
    end subroutine readhist

    subroutine commit_to_mem
      !Take worker history that first points to and store in corresponding work(wid)%first location.
      integer :: styr,stweek,ten,hist_start,hist_end,enweek,myi
      type(hist), pointer :: hp,hp_prev,hpstart !used to point at particular link

      if(.not. associated(first)) return
      !if running the whole sample, then use this one.
        !styr=max(53,work(wc)%gradyr+1) !history starts either in 1987 or 1 year after graduation.
        !hist_start=max(51,work(wc)%gradyr+1) !Initial condition from 1985
        !hist_end=min(79,work(wc)%byr+50) !History ends in 2013
      if(name .EQ. 'time1') then
        styr=max(55,work(wc)%gradyr+1) !history starts either in 1989 or 1 year after graduation.
        hist_start=max(53,work(wc)%gradyr+1) !Initial condition from 1987
        hist_end=min(59,work(wc)%byr+50) !History ends in 1993
      else if(name .EQ. 'time2') then
        styr=max(60,work(wc)%gradyr+1) !history starts either in 1994 or 1 year after graduation.
        hist_start=max(58,work(wc)%gradyr+1) !Initial condition from 1992
        hist_end=min(64,work(wc)%byr+50) !History ends in 1998
      else if(name .EQ. 'time3') then
        styr=max(65,work(wc)%gradyr+1) !history starts either in 1999 or 1 year after graduation.
        hist_start=max(63,work(wc)%gradyr+1) !Initial condition from 1997
        hist_end=min(69,work(wc)%byr+50) !History ends in 2003
     else if(name .EQ. 'time4') then
        styr=max(70,work(wc)%gradyr+1) !history starts either in 2004 or 1 year after graduation.
        hist_start=max(68,work(wc)%gradyr+1) !Initial condition from 2002
        hist_end=min(74,work(wc)%byr+50) !History ends in 2008
     else
        styr=max(75,work(wc)%gradyr+1) !history starts either in 2009 or 1 year after graduation.
        hist_start=max(73,work(wc)%gradyr+1) !Initial condition from 2007
        hist_end=min(79,work(wc)%byr+50) !History ends in 2013
    end if
      
      if(styr>80) stop'start year greater than 80'
      stweek=yrtoweek(styr)
      enweek=yrtoweek(hist_end+1) !this is the last week of the worker's potential emp history
      
      hp=>first
    
      do while(hp%so%start<stweek)
         hp=>hp%next
         if(.not. associated(hp)) then
             print*, 'warning: exiting without committing to mem', first%so%pnr
             write(11,'("warning: exiting without committing to mem",i15)') first%so%pnr
             return
        end if
      end do
      
      if(.not. spcheck(hp)) hp=>nextemp(hp%next) !hp isn't pointing to valid spell. find next one.
      if(.not. associated(hp)) return

      if(hp%so%yr==work(wc)%gradyr+1) then
         hpstart=>hp
      else
         hpstart=>spstart(hp)
      end if
      if(hpstart%so%start<=stweek) then
         hp%so%start=stweek
         hp%so%spellstart=hpstart%so%start
      end if

      !Pad with initial non-employment spell if start date after stweek.
      if(hp%so%start>stweek) then !pad with an unemployment spell
         hp_prev=>prevemp(hp%prev)
         if(associated(hp_prev)) then
            if(hp_prev%so%yr<work(wc)%gradyr+1) hp_prev=>null()
         end if
         if(associated(hp_prev)) then   !need to subtract tenure by 1 because: 
            ten=stweek-hp_prev%so%end-1 !if previous spell ends at the end of week 3 and next spell starts at the beginning of week 4 then tenure =0
            if(ten<-1) then
               if(master) write(6,'("negative tenure",2i10,4i6)') wc,hp%so%pnr,stweek,ten,hp%so%start,hp_prev%so%end
               stop
            end if
            ten=max(0,ten)
         else
            ten=max(0,stweek-yrtoweek(hist_start)-1) !same here you need to subtract it by 1.
         end if
         if(ten<0) then
            write(6,'("negative tenure",2i10,3i6)') wc,hp%so%pnr,stweek,ten,hp%so%start
            write(6,'(i10)') hist_start
            stop
         end if
         sp=>create_iniunemp_spell(stweek,ten,hp%so%start-1,nonemp_endyr(hp)-styr+1,styr)
         call add_worker(work(sp%wid),sp)
         call add_firm(firm(sp%fid),sp)
      end if

      !assemble employment and nonemployment spells.
      call makespells(hp)
      
      !pad end of history with non-employment if last spell ends before the min of 2013 and year where worker is 49.
      if(work(wc)%last%end<enweek-8) then !The worker's spell history is ending before end of either 2013 or 49th year. Pad with non-employment.
          myi=1
          do while(yrtoweek(myi)<work(wc)%last%end+1)
              myi=myi+1
          end do
          myi=myi-1 !this is the start year of the non-emp spell
          sp=>create_termunemp_spell(work(wc)%last%end+1,enweek,hist_end-myi+1,myi)
          mypad=mypad+1
!					write(6,'("Padded worker history with terminal nonemp spell.",4i12)') wc,work(wc)%id,work(wc)%last%end+1,enweek
          call add_worker(work(sp%wid),sp)
          call add_firm(firm(sp%fid),sp)
      end if
    end subroutine commit_to_mem

    recursive subroutine makespells(myhist)
      type(hist), pointer, intent(in) :: myhist
      type(hist), pointer :: myhp,mynext
      call create_emp_spell(myhist,myhp) !create emp spell. return with myhp pointing to last obs of spell.
      mynext=>nextemp(myhp%next)
      if(.not. associated(mynext)) return !we are done
      if(mynext%so%start-myhp%so%end>minunemdur) then !create a non-employment spell
         call create_unemp_spell(myhp,mynext)
      end if
      call makespells(mynext) !make next employment spell
    end subroutine makespells

    subroutine create_emp_spell(myhist,myhp) !this call subject to myhist first observation of proper spell
      type(hist), pointer, intent(in) :: myhist
      type(hist), pointer, intent(out) :: myhp
      type(spell), pointer :: res
      allocate(res)
      res%wid=wc
      res%fid=find_fid(myhist%so%cvr) !the spell fid is the index of the firm in the panel (as distinctly different from firm(fid)%id, which is the CVR number of the firm)
      myhp=>spend(myhist) !make myhp point to last observation in job spell
      allocate(res%wage(myhp%so%yr-myhist%so%yr+1))
      call wagepop(myhist,res%wage) !populate res%wage with wage rate observations			
			res%hours_per_week=sphours(myhist)/spdur(myhist)
      res%yr=myhist%so%yr
      res%strt=myhist%so%start
      res%end=myhp%so%end
      res%ten=max(0,myhist%so%start-max(myhist%so%spellstart,yrtoweek(work(wc)%gradyr+1)))
      call add_worker(work(res%wid),res)
      call add_firm(firm(res%fid),res)
    end subroutine create_emp_spell

    subroutine create_unemp_spell(myhp1,myhp2) !this call subject to myhist first observation of proper spell
      type(hist), pointer, intent(in) :: myhp1,myhp2
      type(spell), pointer :: res
      integer :: dur
      allocate(res)
      res%wid=wc
      res%fid=0
      res%strt=myhp1%so%end+1			
      if(res%strt>=yrtoweek(myhp1%so%yr+1)) then
         res%yr=myhp1%so%yr+1
      else
         res%yr=myhp1%so%yr
      end if
      dur=nonemp_endyr(myhp2)-res%yr+1
      if(dur<1) stop'Fatal(create_unemp_spell): year dur is less than 1.'
      allocate(res%wage(dur))
      res%wage=0.d0
      res%end=max(res%strt,myhp2%so%start-1)
      res%ten=0
      res%hours_per_week=0.d0
      call add_worker(work(res%wid),res)
      call add_firm(firm(res%fid),res)
    end subroutine create_unemp_spell

    subroutine wagepop(myhist,myw) !populate myw with wage observations
      type(hist), pointer, intent(in) :: myhist
      type(hist), pointer :: myhp
      real(4), dimension(:), intent(inout) :: myw
      real(4) :: tothrs
      real(8) :: earn(size(myw)),hours(size(myw))
      integer :: weeks(size(myw))
      integer :: i
      earn=0.d0
      hours=0.d0
      myhp=>myhist
      do
      	 if(myhp%so%yr-myhist%so%yr+1<1 .or. myhp%so%yr-myhist%so%yr+1>size(myw)) stop'wagepop failed'
         earn(myhp%so%yr-myhist%so%yr+1)=earn(myhp%so%yr-myhist%so%yr+1)+myhp%so%earnings !myist%so is still the first spell observation, myhp keeps pointing to the next observation
         hours(myhp%so%yr-myhist%so%yr+1)=hours(myhp%so%yr-myhist%so%yr+1)+myhp%so%hours
				 weeks(myhp%so%yr-myhist%so%yr+1)=myhp%so%end-myhp%so%start+1
         myhp=>myhp%next !update myhp to the next spell observation
         if(.not. associated(myhp)) exit
         if(myhp%so%cvr/=myhp%prev%so%cvr .or. myhp%so%start-myhp%prev%so%end>recalltol) exit !if change firm or if going back to the same firm with a big gap of unemployment then exit
      end do
      do i=1,size(hours)
				 if(weeks(i)<=0 .or. weeks(i)>54) stop'weeks is zero or less or greater than 54.'
         if(hours(i)<=0.d0) stop'hours sum is zero'
				 if(earn(i)<=0.d0) stop'earnings are negative or zero.'
      end do
 !     myw=log(earn/hours)
      myw=log(earn/dble(weeks))
    end subroutine wagepop

    recursive function spend(myhist) result(res) !return pointing to last link of a job spell.
      type(hist), pointer, intent(in) :: myhist
      type(hist), pointer :: res
			res=>null()
			if(.not. associated(myhist)) return
      res=>myhist
      if(associated(myhist%next)) then
         if(myhist%next%so%cvr==myhist%so%cvr .and. myhist%next%so%start-myhist%so%end<=recalltol) res=>spend(myhist%next)
      end if
    end function spend

    recursive function spstart(myhist) result(res) !return pointing to first link of a job spell.
      type(hist), pointer, intent(in) :: myhist
      type(hist), pointer :: res
      res=>myhist
      if(associated(myhist%prev)) then
         if(myhist%prev%so%cvr==myhist%so%cvr .and. myhist%so%start-myhist%prev%so%end<=recalltol) res=>spstart(myhist%prev)
      end if
    end function spstart

    function spcheck(myhist) result(res) !return logical true if spell has more than hours limit
      type(hist), pointer, intent(in) :: myhist
      logical :: res
      real(8) :: dur,hrs,earn
      res=.false.
      if(.not. associated(myhist)) return
      dur=spdur(myhist)
			hrs=sphours(myhist)
			earn=spearn(myhist)
       if(earn/hrs>0.d0) res=.true.
      !if(hrs/dur>5.d0 .and. earn/hrs>1.d0) res=.true.
    end function spcheck

		function spdur(myhist) result(res)
      type(hist), pointer, intent(in) :: myhist
      integer :: res
     	type(hist), pointer :: hp_start,hp_end
			res=0
      if(.not. associated(myhist)) return
     	hp_start=>spstart(myhist)
     	hp_end=>spend(myhist)
     	res=hp_end%so%end-hp_start%so%start+1
    end function spdur
    
    function sphours(myhist) result(res)
   		type(hist), pointer, intent(in) :: myhist
   		real(8) :: res
   		type(hist), pointer :: myhp
   		res=0.d0
   		if(.not. associated(myhist)) return
   		myhp=>spstart(myhist)
   		res=sphours_forward(myhp)
   	end function sphours

    function spearn(myhist) result(res)
   		type(hist), pointer, intent(in) :: myhist
   		real(8) :: res
   		type(hist), pointer :: myhp
   		res=0.d0
   		if(.not. associated(myhist)) return
   		myhp=>spstart(myhist)
   		res=spearn_forward(myhp)
   	end function spearn
     	
    recursive function sphours_forward(myhist) result(res) !count total number of hours in a job spell.
      type(hist), pointer, intent(in) :: myhist
      real(8) :: res
      res=myhist%so%hours
      if(associated(myhist%next)) then
         if(myhist%next%so%cvr==myhist%so%cvr .and. myhist%next%so%start-myhist%so%end<=recalltol) res=res+sphours_forward(myhist%next)
      end if
    end function sphours_forward

    recursive function spearn_forward(myhist) result(res) !count total number of hours in a job spell.
      type(hist), pointer, intent(in) :: myhist
      real(8) :: res
      res=myhist%so%earnings
      if(associated(myhist%next)) then
         if(myhist%next%so%cvr==myhist%so%cvr .and. myhist%next%so%start-myhist%so%end<=recalltol) res=res+spearn_forward(myhist%next)
      end if
    end function spearn_forward

    recursive function sphours_back(myhist) result(res) !count total number of hours in a job spell.
      type(hist), pointer, intent(in) :: myhist
      real(8) :: res
      res=myhist%so%hours
      if(associated(myhist%prev)) then
         if(myhist%prev%so%cvr==myhist%so%cvr .and. myhist%so%start-myhist%prev%so%end<=recalltol) res=res+sphours_back(myhist%prev)
      end if
    end function sphours_back

    function nonemp_endyr(myhist) result(res)
      type(hist), pointer, intent(in) :: myhist
      integer :: res
      if(yrtoweek(myhist%so%yr)<myhist%so%start) then
         res=myhist%so%yr
      else
         res=myhist%so%yr-1
      end if
    end function nonemp_endyr

    recursive function nextemp(myhist) result(res)
      type(hist), intent(in), pointer :: myhist
      type(hist), pointer :: res,temp
      res=>null()
      if(.not. associated(myhist)) return
      if(spcheck(myhist)) then !to count a spell it must have a total of at least 100 hours.
         res=>myhist
      else
      	 temp=>spend(myhist)
         if(associated(temp%next)) res=>nextemp(temp%next)
      end if
    end function nextemp

    recursive function prevemp(myhist) result(res)
      type(hist), intent(in), pointer :: myhist
      type(hist), pointer :: res,temp
      res=>null()
      if(.not. associated(myhist)) return
      if(spcheck(myhist)) then !to count a spell it must have a total of at least 100 hours.
         res=>myhist
      else
      	 temp=>spstart(myhist)
         if(associated(temp%prev)) res=>prevemp(temp%prev)
      end if
    end function prevemp

    function create_iniunemp_spell(mystart,myten,myend,yrdim,myyrstart) result(res)
      integer, intent(in) :: mystart,myten,myend,yrdim,myyrstart
      type(spell), pointer :: res
      allocate(res)
      res%wid=wc !take from global
      res%fid=0
      if(yrdim<1) stop'fatal: yrdim is less than 1.'
      allocate(res%wage(yrdim))
      res%wage=0.d0
      res%yr=myyrstart
      res%strt=mystart
      res%end=myend
      res%ten=myten
    end function create_iniunemp_spell

    function create_termunemp_spell(mystart,myend,yrdim,myyrstart) result(res)
      integer, intent(in) :: mystart,myend,yrdim,myyrstart
      type(spell), pointer :: res
      allocate(res)
      res%wid=wc !take from global
      res%fid=0
      if(yrdim<1) stop'fatal: yrdim is less than 1.'
      allocate(res%wage(yrdim))
      res%wage=0.d0
      res%yr=myyrstart
      res%strt=mystart
      res%end=myend
      res%ten=0
    end function create_termunemp_spell
    
    subroutine readpnr
      integer :: i
      type(pnrobs) :: myobs
      character(20) :: mylabel(8)
      logical :: readagain
      !Do a read through to establish worker dimensions.
      if(allocated(work)) then
         do i=1,dim_I
            if(associated(work(i)%first)) call remove_spell(work(i)%first)
         end do
         deallocate(work)
      end if

      dim_I=0

      if(master) then
         open(10,file=datpath//name//"_pnr.csv",status='old')
         read(10,*) mylabel
         do
            read(10,*,iostat=info) myobs
            if(info<0) exit
            dim_I=dim_I+1
         end do
         close(10)
      end if
      if(master) write(6,'("Reading person data. Number of individuals:",I10)') dim_I
      call mpi_barrier(mpi_comm_world,ierr)
      call mpi_bcast(dim_I,1,mpi_integer,0,mpi_comm_world,ierr) !communicate dim_I to everybody.

      allocate(work(dim_I))

      !Now, populate worker side
      open(10,file=datpath//name//"_pnr.csv",status='old')
      read(10,*) mylabel
      i=0
      do
         readagain=.true.
         do while(readagain)
            read(10,*,iostat=info) myobs
            if(info<0) then
               readagain=.false.
               close(10)
               myobs%pnr=-1
            else
               readagain=.false.
            end if
         end do
         if(myobs%pnr==-1) exit !leave the loop

         i=i+1
         work(i)%id=myobs%pnr
         work(i)%byr=myobs%birthyr-1934
         !          work(i)%startweek=myobs%lagstartspell
         work(i)%edu=myobs%elen
         work(i)%gradyr=myobs%gradyr-1934
         if(myobs%female==1) then
            work(i)%male=.false.
         else
            work(i)%male=.true.
         end if

      end do
      if(i .ne. dim_I) stop'something went wrong (readpnr)'
      close(10)
    end subroutine readpnr

    subroutine readcvr
      integer :: i
      type(cvrobs) :: myobs
      character(20) :: mylabel(3)
      !Do a read through to establish worker and firm dimensions.
      dim_J=0
      if(allocated(firm)) deallocate(firm)
      if(master) then
         open(10,file=datpath//name//"_cvr.csv",status='old')
         read(10,*) mylabel
         do
            read(10,*,iostat=info) myobs
            if(info<0) exit
            dim_J=dim_J+1
         end do
         close(10)
      end if
      call mpi_barrier(mpi_comm_world,ierr)
      call mpi_bcast(dim_J,1,mpi_integer,0,mpi_comm_world,ierr)
      if(master) write(6,'("(readcvr): Master has determined and broadcast a firm dimension of ",i15)') dim_J
      allocate(firm(0:dim_J))
      firm(0)%id=0
      !Now, populate firm side
      open(10,file=datpath//name//"_cvr.csv",status='old')
      read(10,*) mylabel
      do i=1,dim_J
         read(10,*) myobs
         firm(i)%id=myobs%cvr !firm%id holds the cvr number
         firm(i)%sector=myobs%sector
         firm(i)%industry=myobs%industry
         firm(i)%bweek=myobs%bweek
         firm(i)%s=firm(i)%sector*12+firm(i)%industry+1 !Assumes 12 indu cat and 3 sector cat, where 0 is not classified. It is possible that certain categories are empty.
      end do
      close(10)
    end subroutine readcvr

    function find_fid(inp2) result(res) !return with lowest res such that inp1(res)>=inp2
                                        !inp2 is the firm CVR number. function returns index s.t. firm(res)%fid=inp2
      integer, intent(in) :: inp2
      integer :: res
      integer :: dim,lb,ub
      if(inp2>firm(dim_J)%id .or. inp2<firm(1)%id) then
         write(6,'("I am process", i6)') myid
         write(6,'(4i15)') dim_J, firm(dim_J)%id, firm(1)%id, inp2
         stop'find failed. Outside bounds.'
      end if
      if(inp2==firm(1)%id) then
         res=1
         return
      end if
      lb=1
      ub=dim_J
      do while(ub-lb>1)
         res=(ub+lb)/2
         if(firm(res)%id<inp2) lb=res
         if(firm(res)%id>=inp2) ub=res
      end do
      
      res=ub
      if(firm(res)%id == inp2) return !design so that function only returns if proper solution is found

      write(6,'("(find_fid): find failed.",3i15)') firm(res)%id,inp2,res
      stop
    end function find_fid
  end subroutine readdat
end module datbyread