module class_spell
  implicit none !all objects in module inherit implicit none
  type spell
     integer(4) :: fid !Firm ID !Can go and look at characteristics data with IDs.
     integer(4) :: wid !Worker ID - Can go and look at characteristics with worker ID.
     real(4) :: hours_per_week
     real(4), pointer :: wage(:)   !can have multiple observations for each year the spell stretches over.
     integer(2) :: strt !start week
     integer(2) :: end !end week
     integer(2) :: ten !tenure at wage(1) - measured in weeks.
     integer(1) :: yr !yr of spell start
     type(spell), pointer :: prev=>null(), next => null()
     type(spell), pointer :: f_prev=>null(), f_next=>null() !to navigate firm panels.
  end type spell
  type(spell), pointer :: this !current spell
contains
  subroutine addspell !add a spell to this, and make this point to the newly added spell.
    type(spell), pointer :: new
    allocate(new)
    new%prev=>this !create backward link
    this%next=>new !create forward link
    nullify(new%next)
    this=>new
  end subroutine addspell

  subroutine newspell !create a new spell, and make this point to it.
    type(spell), pointer :: new
    allocate(new)
    this=>new
  end subroutine newspell

  recursive subroutine remove_spell(mythis) !Removes this and all elements following it.
    type(spell), pointer :: mythis, next
    next => mythis%next
    do while(associated(next))
       call remove_spell(next)
    end do
    if(associated(mythis%wage)) deallocate(mythis%wage) !step in and deallocate wage inside spell.
    deallocate(mythis)
  end subroutine remove_spell

  recursive function persyr_firm(myspell) result(res)
    type(spell), pointer, intent(in) :: myspell
    integer(4) :: res
    res=0
    if(.not. associated(myspell)) then
       return
    else
       if(associated(myspell%wage)) res=size(myspell%wage)
    end if
    if(associated(myspell%f_next)) res=res+persyr_firm(myspell%f_next)
  end function persyr_firm
  
  recursive subroutine life_firm(myspell,myminwk,mymaxwk,res)
    type(spell), pointer, intent(in) :: myspell
    integer, intent(inout) :: myminwk, mymaxwk
    integer, intent(out) :: res(2)
    
    if(.not. associated(myspell)) then
        res(1)=myminwk
        res(2)=mymaxwk
        return
    else if(associated(myspell) .and. .not. associated(myspell%f_prev)) then
        res(1)=myspell%strt
        res(2)=myspell%end
        myminwk=res(1)
        mymaxwk=res(2)
    else
        if(myminwk>myspell%strt) then
            res(1)=myspell%strt
            myminwk=res(1)
        end if
        if(mymaxwk<myspell%end) then
            res(2)=myspell%end
            mymaxwk=res(2)
        end if
    end if
    if(associated(myspell%f_next)) call life_firm(myspell%f_next,myminwk,mymaxwk,res)
  end subroutine life_firm

  recursive subroutine find_t(myspell,myt)
    type(spell), pointer, intent(in) :: myspell
    integer, intent(in) :: myt
    this=>myspell
    if(myspell%end .lt. myt .and. associated(myspell%next)) call find_t(myspell%next,myt)
  end subroutine find_t

  function dur(mythis) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer(4) :: res
    res=mythis%end-mythis%strt+1
  end function dur
end module class_spell

module panels
  use class_spell
  type workerlist
        real(8) :: akm_wage
     integer(8) :: id
     integer(4) :: entry
     integer(4) :: byr !birth year
     integer(4) :: gradyr !graduation yr
     integer(1) :: edu
     integer(1) :: k !worker type (for simulation)
     integer(1) :: z !worker z
     logical :: male
     logical :: native
		 logical :: akm_defined
     type(spell), pointer :: first=>null(),last=>null()
  end type workerlist

  type firmlist
     integer(1) :: industry !industry
     integer(1) :: sector !sector:0=no info, 1=public, 2=private
		 integer(1) :: s !ind x sec
     integer :: l !firm type (for simulation)
     integer(8) :: id
     integer(4) :: bweek !firm's birth week
     type(spell), pointer :: first=>null(),last=>null()
  end type firmlist
contains
  function start(this,mythis)
    type(workerlist) :: this
    type(spell), pointer :: mythis
    integer(4) :: start
    if(associated(mythis%prev)) then
        start=mythis%prev%end+1
    else
        start=this%entry
    end if
  end function start
  
  subroutine add_worker(this,myspell)
    type(workerlist) :: this
    type(spell), pointer, intent(inout) :: myspell
    if(associated(this%last)) then
       myspell%prev=>this%last
       this%last%next=>myspell
       this%last=>myspell
    else
       this%first=>myspell
       this%last=>myspell
    end if
  end subroutine add_worker

  subroutine add_firm(this,myspell)
    type(firmlist) :: this
    type(spell), pointer, intent(inout) :: myspell
    if(associated(this%last)) then
       myspell%f_prev=>this%last
       this%last%f_next=>myspell
       this%last=>myspell
    else
       this%first=>myspell
       this%last=>myspell
    end if
  end subroutine add_firm

  function persyr(this) result(res)
    type(firmlist) :: this
    integer(4) :: res
    res=persyr_firm(this%first)
  end function persyr
  
  function firmlife(this) result(res)
    type(firmlist) :: this
    integer :: res(2), minwk,maxwk
    minwk=10000
    maxwk=0
    call life_firm(this%first,minwk,maxwk,res)
  end function firmlife

  recursive subroutine write_spell(this,mych,mythis) !write out a spell to output channel mych
    type(workerlist) :: this
    integer, intent(in) :: mych
    type(spell), pointer, intent(in) :: mythis
    integer :: i
    write(mych,'(6i10,f15.5)') mythis%wid, mythis%fid, mythis%yr, mythis%ten, mythis%strt,mythis%end,mythis%hours_per_week
    write(mych,'(a15,<size(mythis%wage)>i15)') " ",(i,i=1,size(mythis%wage))
    write(mych,'(a15,<size(mythis%wage)>f15.4)') "Wage",(mythis%wage(i),i=1,size(mythis%wage))
    write(mych,'(" ")')
    if(associated(mythis%next)) call write_spell(this,mych,mythis%next)
  end subroutine write_spell
end module panels

module estim
  use linmod
  use class_spell
  use panels
  use modelpar_class
  implicit none
  type(modelpar) :: ep !Here, hold the estimated model parameters.
  type(modelpar) :: truth
  real(8),allocatable :: pprobw(:,:)  ! estimated posterior probabilities of worker types from step 2
  integer(1),allocatable :: firmtype(:)  !firm type assigned by max posterior prob
   integer(1),allocatable :: worktype(:)  !worker type assigned by kmeans
  integer,allocatable :: firmsize(:) !number of spells in firm.
  integer(4) :: itcount,globi,yrtoweek(80)

  !The major data panels
  type(workerlist), allocatable :: work(:) !these two structures hold the spell data
  type(firmlist), allocatable :: firm(:)
  integer :: dim_I,dim_J
  
  !procedure overloads.
  interface find
     module procedure find_d,find_i
  end interface find
  interface fdensity
     module procedure fdensity_i4,fdensity_i1
  end interface fdensity  
contains
  subroutine estim_ini(ck,cl,dim_I,dim_J,myhetdim) !Estimation is initialized.
    integer, intent(in) :: ck,cl,dim_I,dim_J,myhetdim
    ep=new_modelpar(ck,cl)
    allocate(pprobw(dim_I,ck),firmtype(0:dim_J),firmsize(1:dim_J),worktype(1:dim_I))
  end subroutine estim_ini

  subroutine estim_term
    deallocate(pprobw,firmtype,firmsize,worktype)
    call modelpar_deallocate(ep)
  end subroutine estim_term
  
  subroutine amatmap(myindex,myhetdim,mytyp,myl) !inverting index into type and l for A matrix.
    integer, intent(in) :: myindex, myhetdim 
    integer, intent(out) :: mytyp,myl
    integer :: i
    do i=1,hetdim
      if(myindex<=i*ep%l .and. myindex > (i-1)*ep%l) then
        mytyp=i
        if(i==1) then
            myl=myindex
        else
            myl=myindex-(i-1)*ep%l
            if(myl<0) stop'amatmap index has error'
        end if
      end if   
    end do
   end subroutine amatmap
  
  function hetmap(mysp,myj) !maps into [1,8] integers. [1,4] exp in short and [5,8] in long tenure
    type(spell), pointer :: mysp
    integer, intent(in) :: myj !wage observation number within spell
    integer :: hetmap,myexp
    !Dimensions of heterogeneity: experience (4), tenure (2) (8 groups)
    !Potential experience mapping is set
    !!Use this two lines below for simulation
 !   hetmap=1
 !   return
    myexp=mysp%yr+myj-work(mysp%wid)%gradyr !we assume work history starts one year after grad year
    if(myexp<=5) hetmap=1
    if(myexp>5 .and. myexp<=10) hetmap=2
    if(myexp>10 .and. myexp<=15) hetmap=3
    if(myexp>15) hetmap=4

    !Tenure mapping is set
    
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) hetmap=hetmap+4 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) hetmap=hetmap+4 ! nonemployed tenure. distinguish between first 6 months and everything else. 
  
  end function hetmap
  
  function sdhetmap(mysp,myj) !maps into [1,12] integers. [1,4] exp in 0-1 yr of tenure and [5,8] in 1-2 yr of tenure and [9,12] in 2+ tenuer
    type(spell), pointer :: mysp
    integer, intent(in) :: myj !wage observation number within spell
    integer :: sdhetmap,myexp, myten
    !Dimensions of heterogeneity: experience (4), tenure (2) (8 groups)
    !Potential experience mapping is set
    !!Use this two lines below for simulation
 !   sdhetmap=1
 !   return
    myexp=mysp%yr+myj-work(mysp%wid)%gradyr !we assume work history starts one year after grad year
    if(myexp<=5) sdhetmap=1
    if(myexp>5 .and. myexp<=10) sdhetmap=2
    if(myexp>10 .and. myexp<=15) sdhetmap=3
    if(myexp>15) sdhetmap=4

    !Tenure mapping is set
    myten=yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten
    if(myten>=52 .and. myten<100 .and. mysp%fid>0) sdhetmap=sdhetmap+4 ! employed tenure. distinguish between 0-1, 1-2 and 2+ years.
    if(myten>=100 .and. mysp%fid>0) sdhetmap=sdhetmap+8 ! employed tenure. distinguish between 0-1, 1-2 and 2+ years.
    if(mysp%fid==0) stop'compute SD when unemployed' ! we only read in employment spells here
  
  end function sdhetmap

  function inihetmap(mysp) result(res) !maps into [1,8] integers. [1,4] in each tenure type
    type(spell), pointer :: mysp
    integer :: res
    integer :: myexp
    !!Use these two lines below for simulation
 !   res=1
 !   return
    !Dimensions of heterogeneity: exp (4) - only for first time group, time (9), tenure (2) (24 groups)
    !Age mapping is set	
    myexp=mysp%yr-work(mysp%wid)%gradyr
    if(myexp<=5) res=1
    if(myexp>5 .and. myexp<=10) res=2
    if(myexp>10 .and. myexp<=15) res=3
    if(myexp>15) res=4

    !Tenure mapping is set
    if(yrtoweek(mysp%yr)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+4 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+4 ! nonemployed tenure. distinguish between first 6 months and everything else.

  end function inihetmap

  function mobhetmap(mysp,myj) result(res)
    type(spell), pointer :: mysp
    integer, intent(in) :: myj
    integer :: res,myexp
    !!Use these two lines below for simulation
!    res=1
!    return
    myexp=mysp%yr+myj-work(mysp%wid)%gradyr
    if(myexp<=5) res=1
    if(myexp>5 .and. myexp<=10) res=2
    if(myexp>10 .and. myexp<=15) res=3
    if(myexp>15) res=4


    !Tenure mapping is set
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+4 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+4 ! nonemployed tenure. distinguish between first 6 months and everything else.
    
  end function mobhetmap

  function zhetmap(mywork) result(res)
    type(workerlist), intent(in) :: mywork
    integer ::res
    !!Use this two lines below for simulation
!    res=1
!    return
    if(mywork%edu<12) res=1
    if(mywork%edu==12) res=2
    if(mywork%edu>12) res=3
    if(mywork%male) res=res+3
  end function zhetmap

  function find_d(in1,in2) result(res)
    real(8), dimension(:), intent(in) :: in1
    real(8), intent(in) :: in2
    integer :: res
    integer :: dim,lb,ub
    if(in2<=in1(1)) then
       res=1
       return
    end if
    dim=size(in1)
    if(in2>in1(dim)) then
       res=0
       return
    end if
    lb=1
    ub=dim
    do while(ub-lb>1)
       res=(ub+lb)/2
       if(in1(res)<in2) lb=res
       if(in1(res)>=in2) ub=res
    end do
    res=ub
    if(in1(res-1)>=in2 .or. in1(res)<in2) stop'find failed.'
  end function find_d

  function find_i(in1,in2) result(res)
    integer, dimension(:), intent(in) :: in1
    integer, intent(in) :: in2
    integer :: res
    integer :: dim,lb,ub
    if(in2<=in1(1)) then
       res=1
       return
    end if
    dim=size(in1)
    if(in2>in1(dim)) then
       res=0
       return
    end if
    lb=1
    ub=dim
    do while(ub-lb>1)
       res=(ub+lb)/2
       if(in1(res)<in2) lb=res
       if(in1(res)>=in2) ub=res
    end do
    res=ub
    if(in1(res-1)>=in2 .or. in1(res)<in2) stop'find failed.'
  end function find_i

  function fdensity_onew_ar(mythis,k,ll,t)
    type(spell), pointer, intent(in) :: mythis
    integer, intent(in) :: k, ll, t !t is time period of wage observation
    integer :: typ, typlag, sdtyp
    real(8) :: w, lastwage, lastmuw
    real(8) :: fdensity_onew_ar
    real(8) :: pi = 3.1415926535897932d0
    fdensity_onew_ar = 0.d0     ! Initialize density
    if(ll==0) return
    w=mythis%wage(t)
    typ=hetmap(mythis,t)
    sdtyp=sdhetmap(mythis,t)
    lastwage=mythis%wage(t-1)
    typlag=hetmap(mythis,t-1)
    lastmuw=ep%muw(k,ll,typlag)
    fdensity_onew_ar = -log(ep%sigmaw(k,ll,sdtyp))-0.5d0*(log(2.d0)+log(pi))-0.5d0*(((w-ep%muw(k,ll,typ)-ep%rho*(lastwage-lastmuw))/ep%sigmaw(k,ll,sdtyp))**2)
    if(fdensity_onew_ar<-huge(fdensity_onew_ar)) then
       write(6,'("fdensity_ar is ill defined",2i10,2f15.5,5i10)') 1,mythis%wid,ep%muw(k,ll,typ),w,typ,sdtyp,k,ll,t
       fdensity_onew_ar=-huge(fdensity_onew_ar)
    end if
    if(isnan(fdensity_onew_ar)) then
!       print *,2,mythis%wid,w,typ,k,ll,t
       fdensity_onew_ar=-huge(fdensity_onew_ar)
    end if
  end function fdensity_onew_ar
  
  function fdensity_onew_arf(mythis,myprev, k,ll,t) !for move between jobs and AR is carried over. But we don't use this anymore. 
    type(spell), pointer, intent(in) :: mythis, myprev
    integer, intent(in) :: k, ll, t !t is time period of wage observation
    integer :: typ, typlag, lastperiod
    real(8) :: w, lastwage, lastmuw, jlag
    real(8) :: fdensity_onew_arf
    real(8) :: pi = 3.1415926535897932d0
    fdensity_onew_arf = 0.d0     ! Initialize density
    if(ll==0) return
    myprev=mythis%prev
    w = mythis%wage(t)
    typ=hetmap(mythis,t)
    jlag = myprev%fid
    if(firmtype(jlag)>0) then
        lastperiod=size(myprev%wage)
        lastwage=myprev%wage(lastperiod)
        typlag=hetmap(myprev,lastperiod)
        lastmuw=ep%muw(k,firmtype(jlag),typlag)
    else
        lastwage=0.d0
        lastmuw=0.d0
    end if
    fdensity_onew_arf = -log(ep%sigmaw(k,ll,typ))-0.5d0*(log(2.d0)+log(pi))-0.5d0*(((w-ep%muw(k,ll,typ)-ep%rho*(lastwage-lastmuw))/ep%sigmaw(k,ll,typ))**2)
    if(fdensity_onew_arf<-huge(fdensity_onew_arf)) then
       write(6,'("fdensity_arf is ill defined",2i10,2f15.5,4i10)') 1,mythis%wid,ep%muw(k,ll,typ),w,typ,k,ll,t
       fdensity_onew_arf=-huge(fdensity_onew_arf)
    end if
    if(isnan(fdensity_onew_arf)) then
!       print *,2,mythis%wid,w,typ,k,ll,t
       fdensity_onew_arf=-huge(fdensity_onew_arf)
    end if
  end function fdensity_onew_arf
  
  function fdensity_onew(mythis,k,ll,t)
    type(spell), pointer, intent(in) :: mythis
    integer, intent(in) :: k, ll, t
    integer :: typ
    real(8) :: w
    real(8) :: fdensity_onew
    real(8) :: pi = 3.1415926535897932d0
    fdensity_onew = 0.d0     ! Initialize density
    if(ll==0) return
    w = mythis%wage(t)
    typ=hetmap(mythis,t)
    fdensity_onew = -log(ep%sigmaw(k,ll,typ))-0.5d0*(log(2.d0)+log(pi))-0.5d0*(((w-ep%muw(k,ll,typ))/ep%sigmaw(k,ll,typ))**2)
    if(fdensity_onew<-huge(fdensity_onew)) then
       write(6,'("fdensity is ill defined",2i10,2f15.5,4i10)') 1,mythis%wid,ep%muw(k,ll,typ),w,typ,k,ll,t
       fdensity_onew=-huge(fdensity_onew)
    end if
    if(isnan(fdensity_onew)) then
!       print *,2,mythis%wid,w,typ,k,ll,t
       fdensity_onew=-huge(fdensity_onew)
    end if
  end function fdensity_onew

  function fdensity_i4(mythis,k,ll) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer(4), intent(in) :: k, ll
    real(8) :: res
    integer :: t
    res = 0.d0  ! Initialize density
    if(ll==0) return
    do t = 1, size(mythis%wage)   !do this for all periods in a given spell of each match
      if(t==1) then
        res = res + fdensity_onew(mythis,k,ll,t)  
      else  !for subsequent wages within a spell
        res = res + fdensity_onew_ar(mythis,k,ll,t) 
      end if
    end do
    if(res<-huge(res)) res=-huge(res)
    if(isnan(res)) res=-huge(res)
  end function fdensity_i4

  function fdensity_i1(mythis,k,ll) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer(4), intent(in) :: k
    integer(1), intent(in) :: ll
    real(8) :: res
    res=fdensity_i4(mythis,k,int(ll))
  end function fdensity_i1

  function fdensity_mat(mythis)
    type(spell), pointer, intent(in) :: mythis
    real(8) :: fdensity_mat(ep%k,ep%l)
    integer :: ll, k

    fdensity_mat = 0.d0 ! Initialize density
    do k = 1, ep%k
       do ll = 1, ep%l
          fdensity_mat(k,ll) = fdensity(mythis,k,ll)
       end do
    end do
  end function fdensity_mat
end module estim