module modelpar_class !contains model parameters and objects and operations on them.
  implicit none
  integer, parameter :: hetdim=8    !wage parameter x het dim 
  integer, parameter :: inihetdim=8 !m dist x het dim 
  integer, parameter :: mobhetdim=8 !mob parameter x het dim 
  integer, parameter :: sdhetdim=12 !sd parameter x het dim 
  integer, parameter :: zhetdim=6   !z dim.
  integer, parameter :: shetdim=36   !s dim (number of industries)

  !integer, parameter :: hetdim=1    !wage parameter x het dim 
  !integer, parameter :: inihetdim=1 !m dist x het dim 
  !integer, parameter :: mobhetdim=1 !mob parameter x het dim 
  !integer, parameter :: sdhetdim=1 !sd parameter x het dim 
  !integer, parameter :: zhetdim=2   !z dim.
  !integer, parameter :: shetdim=2   !s dim (number of industries)

  type modelpar
     integer :: k,l
     real(8),allocatable :: muw(:,:,:),sigmaw(:,:,:) !worker and firm type specific (unobservables) wage mean and std dev.
     real(8),allocatable :: gam(:,:,:) !match surplus means by firm and worker unobservables: worker type specific ranks.
     real(8),allocatable :: lam(:,:,:) !(k,l) specific offer arrival rate.
     real(8),allocatable :: m(:,:,:) !initial distribution m(k,l|x) (x conditional)
     real(8),allocatable :: q(:) !distribution of firm types
     real(8),allocatable :: p(:) !distribution of worker types
     real(8),allocatable :: del(:,:,:) !(k,l,x) conditional layoff rate.
     real(8),allocatable :: psi(:,:,:) !probability that a (k,x) individual finds a l' job.
     real(8),allocatable :: MM(:,:,:,:) !M(k,l,ll,x) prob that type k worker moves from l to ll given x
     real(8),allocatable :: Mbar(:,:,:) !\bar{M}(k,l,x) prob that type k worker stays with l given x
     real(8),allocatable :: piw(:,:) !pi_w(z|k)
     real(8),allocatable :: pif(:,:)	!pi_f(s|l). s is the industry x sector indicator. by default s=0 is a non-category for l=0.
     																	!pif needs to handle l=0 matches and so dim of pif runs pif(0:shetdim,0:myl), where pif(0,0)=1.
		 real(8),allocatable :: pz(:) !marginal distribution of z
		 real(8),allocatable :: ps(:) !marginal distribution of s
     real(8) :: rho !AR(1) parameter
  end type modelpar
contains
  function capM(k,l,ll,x,mypar) result(res) !transition probability of moving from l to ll given k.
    integer, intent(in) :: l,ll,k,x
    type(modelpar), intent(in) :: mypar
    real(8) :: res
    if(l==0 .and. ll==0) then !M_{k00}(x)=0 by definition
       res=0.0d0
    else if(l==0) then !Move from unemployment to emp.
       res=mypar%psi(k,ll,x)
    else if (ll==0) then !layoff
       res=mypar%del(k,l,x)
    else
    	 if(mypar%gam(k,l,x)+mypar%gam(k,ll,x)>0.d0) then
             res=mypar%lam(k,ll,x)*mypar%gam(k,ll,x)/(mypar%gam(k,l,x)+mypar%gam(k,ll,x))
         else
             res=0.5d0*mypar%lam(k,ll,x)
        end if
    end if
  end function capM
  
  subroutine M_update(par)
    type(modelpar), intent(inout) :: par
    integer :: l,ll,k,x
    do x=1,mobhetdim
       do k=1,par%k
          do l=0,par%l
             do ll=0,par%l
                par%MM(k,l,ll,x)=capM(k,l,ll,x,par)
             end do
             par%mbar(k,l,x)=1.d0-sum(par%MM(k,l,:,x))
             !if mbar<=0, set it to a small numner and rescale par%MM too
             if(par%mbar(k,l,x)<=0.d0) then
                par%mbar(k,l,x)=0.000001d0
                par%MM(k,l,:,x)=par%MM(k,l,:,x)/sum(par%MM(k,l,:,x))*0.999999d0
            end if
          end do
       end do
    end do
  end subroutine M_update

!  subroutine steadystate(par) 
!    type(modelpar), intent(inout) :: par
!    integer :: j,jj,i,xx !counters
!    real(8) :: amat(0:par%l,0:par%l),bmat(0:par%l)
!    integer :: pivot(0:par%l),info
!    do xx=1,mobhetdim
!       do i=1,par%k
!          do j=0,par%l
!             do jj=0,par%l
!                amat(j,jj)=1.0d0-par%MM(i,jj,j,xx) !mym(i,jj,j) !this evaluates prob of transition from jj to j given i
!                if(j==jj) amat(j,jj)=amat(j,jj)+1.d0-par%Mbar(i,j,xx) !mym(i,j,0,sim_cl) !this sum of transitions from j to jj where jj runs from 1 to sim_cl
!             end do
!          end do
!          bmat=1.d0 !remember etak is specified as CDF, and we want pdf here.
!          call dgesv(par%l+1,1,amat,par%l+1,pivot,bmat,par%l+1,info)
!          par%m(i,0:par%l,xx)=bmat
!       end do
!    end do
!  end subroutine steadystate
  
  function new_modelpar(myk,myl) result(res)
    integer, intent(in) :: myk,myl
    type(modelpar) :: res
    allocate(res%muw(myk,myl,hetdim),res%sigmaw(myk,myl,sdhetdim),res%gam(myk,myl,mobhetdim),res%lam(myk,myl,mobhetdim),&
         &res%m(myk,0:myl,inihetdim),res%q(0:myl),res%p(myk),res%del(myk,myl,mobhetdim),&
         &res%psi(myk,myl,mobhetdim),res%MM(myk,0:myl,0:myl,mobhetdim),res%Mbar(myk,0:myl,mobhetdim),&
         &res%piw(zhetdim,myk),res%pif(0:shetdim,0:myl),res%pz(zhetdim),res%ps(shetdim))
    res%muw   =0.d0
    res%sigmaw=0.d0
    res%gam   =0.d0
    res%lam   =0.d0
    res%m     =0.d0
    res%piw   =0.d0
    res%pif   =0.d0
    res%q     =0.d0
    res%p			=0.d0
    res%del   =0.d0
    res%psi   =0.d0
    res%M     =0.d0
    res%Mbar  =0.d0
    res%q(0)  =1.d0 !unemp group. Not updated. By default one.
    res%k     =myk
    res%l     =myl
    res%rho   =0.d0
    res%pz		=0.d0
    res%ps		=0.d0
  end function new_modelpar

  subroutine modelpar_deallocate(mypar)
    type(modelpar), intent(inout) :: mypar
    if(allocated(mypar%muw)) then
       deallocate(mypar%muw,mypar%sigmaw,mypar%gam,mypar%lam,mypar%m,mypar%p,mypar%q,mypar%del,&
         &mypar%psi,mypar%MM,mypar%Mbar,mypar%piw,mypar%pif,mypar%pz,mypar%ps)
    end if
  end subroutine modelpar_deallocate

  subroutine writemodelpar(mypar,ch,labels,order) !writes mypar to channel ch
    type(modelpar), intent(in) :: mypar
    integer, intent(in) :: ch
    logical, intent(in) :: labels !if true, print labels
    logical, intent(in), optional :: order
    integer :: ii,jj,myk,myl,tt
    integer, allocatable :: kmap(:),lmap(:)
    real(8), allocatable :: muvec(:)
    real(8) :: minwage
    myk=size(mypar%muw,dim=1)
    myl=size(mypar%muw,dim=2)
    allocate(kmap(myk),lmap(0:myl))
    do ii=1,myk
    	kmap(ii)=ii
    end do
    do jj=0,myl
    	lmap(jj)=jj
    end do
    
		!if order, produce ordering by simple averaging on muw
		if(present(order)) then
			if(order) then
				allocate(muvec(myk))
				do ii=1,myk
					muvec(ii)=sum(mypar%muw(ii,:,1))/dble(myl)
				end do

				!now do brute force ascending order
				do tt=1,myk
					minwage=huge(minwage)
					do ii=1,myk
						if(muvec(ii) .lt. minwage) then
							minwage=muvec(ii)
							kmap(tt)=ii
						end if
					end do
					muvec(kmap(tt))=huge(minwage)
				end do

				deallocate(muvec)
				allocate(muvec(myl))
				do jj=1,myl
					muvec(jj)=sum(mypar%muw(:,jj,1))/dble(myk)
				end do

				!now do brute force ascending order
				do tt=1,myl
					minwage=huge(minwage)
					do jj=1,myl
						if(muvec(jj) .lt. minwage) then
							minwage=muvec(jj)
							lmap(tt)=jj
						end if
					end do
					muvec(lmap(tt))=huge(minwage)
				end do
			end if !end of order if
		end if !end of present order if

		write(ch,'(<myk>i5)') kmap
		write(ch,'(<myl+1>i5)') lmap
		
    if(labels) then
      write(ch,'("K=",i3,"  L=",i3)') myk,myl
    else
      write(ch,'(i4)') myk,myl
    end if
    do tt=1,hetdim
      if(labels) write(ch,'("muw", i2)') tt
      do ii=1,myk
        write(ch,'(<myl>f15.8)') (mypar%muw(kmap(ii),lmap(jj),tt), jj=1,myl)
      end do
    end do
    do tt=1,sdhetdim
      if(labels) write(ch,'("sigmaw", i2)') tt
      do ii=1,myk
        write(ch,'(<myl>f15.8)') (mypar%sigmaw(kmap(ii),lmap(jj),tt), jj=1,myl)
      end do
    end do

    do tt=1,mobhetdim
      if(labels) write(ch,'("gam", i2)') tt
      do ii=1,myk
        write(ch,'(<myl+1>f15.8)') (mypar%gam(kmap(ii),lmap(jj),tt), jj=1,myl)
      end do
    end do

    do tt=1,mobhetdim
      if(labels) write(ch,'("lam", i2)') tt
      do ii=1,myk
        write(ch,'(<myl+1>f15.8)') (mypar%lam(kmap(ii),lmap(jj),tt), jj=1,myl)
      end do
    end do

    do tt=1,mobhetdim
      if(labels) write(ch,'("psi", i2)') tt
      do ii=1,myk
        write(ch,'(<myl+1>f15.8)') (mypar%psi(kmap(ii),lmap(jj),tt), jj=1,myl)
      end do
    end do

    do tt=1,mobhetdim
      if(labels) write(ch,'("delta", i2)') tt
      do ii=1,myk
        write(ch,'(<myl+1>f15.8)') (mypar%del(kmap(ii),lmap(jj),tt), jj=1,myl)
      end do
    end do
    
    if(labels) write(ch,'("piw(z,k)")')
    write(ch,'(<myk>f15.8)') ((mypar%piw(tt,kmap(ii)),ii=1,myk),tt=1,zhetdim)

    if(labels) write(ch,'("pif(s,l) - where l=0 is omitted.")')
    write(ch,'(<myl>f15.8)') ((mypar%pif(tt,lmap(jj)),jj=1,myl),tt=1,shetdim)

    do tt=1,inihetdim
      if(labels) write(ch,'("m(:,:,x)", i2)') tt
      do ii=1,myk
        write(ch,'(<myl+1>f15.8)') (mypar%m(kmap(ii),lmap(jj),tt), jj=0,myl)
      end do
    end do

    if(labels) write(ch,'("q")')
    write(ch,'(f15.8)') (mypar%q(lmap(jj)),jj=1,myl)

    if(labels) write(ch,'("p")')
    write(ch,'(f15.8)') (mypar%p(kmap(ii)),ii=1,myk)

    if(labels) write(ch,'("pz")')
    write(ch,'(f15.8)') (mypar%pz(jj),jj=1,zhetdim)

    if(labels) write(ch,'("ps")')
    write(ch,'(f15.8)') (mypar%ps(jj),jj=1,shetdim)
    
    if(labels) write(ch,'("rho")')
    write(ch,'(f15.8)') (mypar%rho)

    do tt=1,mobhetdim
       do myk=1,mypar%k
          if(labels) write(ch,'("MM(k,:,:,x)", "k=",i2, "x=",i2)') myk,tt
          do ii=0,mypar%l
             write(ch,'(<myl+1>f15.8)') (mypar%MM(kmap(myk),lmap(ii),lmap(jj),tt), jj=0,mypar%l)
          end do
       end do
    end do

    do tt=1,mobhetdim
       if(labels) write(ch,'("Mbar(:,:,x)", "x=",i2)') tt
       do myk=1,mypar%k
          write(ch,'(<myl+1>f15.8)') (mypar%Mbar(kmap(myk),lmap(jj),tt), jj=0,mypar%l)
       end do
    end do
  end subroutine writemodelpar

  function modpar_dist(in1,in2)
    type(modelpar), intent(in) :: in1,in2
    real(8) :: modpar_dist
    if(size(in1%muw) .ne. size(in2%muw)) then
       write(6,'("Fatal(modpar_dist): in1 and in2 do not have same dimensions.")')
       stop
    end if
    modpar_dist=sum(sum((in1%muw-in2%muw)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%sigmaw-in2%sigmaw)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%gam-in2%gam)**2,dim=1))
    modpar_dist=modpar_dist+sum((in1%lam-in2%lam)**2)
    modpar_dist=modpar_dist+(in1%rho-in2%rho)**2
    modpar_dist=modpar_dist+sum(sum((in1%m-in2%m)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%piw-in2%piw)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%pif-in2%pif)**2,dim=1))
  end function modpar_dist

  function modparadd(mp1,mp2) result(res)
    type(modelpar), intent(in) :: mp1,mp2
    type(modelpar) :: res
    res=new_modelpar(mp1%k,mp1%l)
    res%muw   =mp1%muw   +mp2%muw
    res%sigmaw=mp1%sigmaw+mp2%sigmaw
    res%gam   =mp1%gam   +mp2%gam
    res%lam   =mp1%lam   +mp2%lam
    res%rho   =mp1%rho   +mp2%rho
  end function modparadd

  function modpardiv(mp1,xx) result(res)
    type(modelpar), intent(in) :: mp1
    real(8),intent(in) :: xx
    type(modelpar) :: res
    res=new_modelpar(mp1%k,mp1%l)
    res%muw   =mp1%muw/xx
    res%sigmaw=mp1%sigmaw/xx
    res%gam   =mp1%gam/xx
    res%lam   =mp1%lam/xx
    res%rho   =mp1%rho/xx
  end function modpardiv
  
  function inttochar(dim,x) result(res)
	implicit none
	integer, intent(in) :: dim,x
	character(dim) :: res
	integer :: rem,i,dig
	character(1) :: num
	rem=x
	res=""
	do i=dim-1,0,-1
		dig=int(dble(rem)/dble(10.d0**i))
		write(num,'(i1)') dig
		res=trim(res)//num
		rem=rem-int(dble(dig)*(10.d0**i))
	end do
end function inttochar
end module modelpar_class
