module datbysim
  !contains simulation specific parameters as well as simulated data structure
  use cem
  use gauscdf
  implicit none !all objects in module inherit implicit none.
  integer, parameter :: mydim_I  = 1000000 !# of workers.
  integer, parameter :: mydim_J  =  100000 !# of firms.
  integer, parameter :: sim_T    =     520 !# of periods.
  integer, parameter :: sim_ck   =       4 !# of worker types.
  integer, parameter :: sim_cl   =       4 !# of firm types.
  integer, parameter :: yr       =      52 !number of periods in a year.
  integer, parameter :: sim_zdim =       2 !number of observable worker types
  integer, parameter :: sim_sdim =       2 !number of observable firm types
  
  type(modelpar) :: par !globally set collection of model parameters. Must be initialized.

  real(8) :: zprob !prob that worker is z=1.
  real(8) :: sprob !prob that firm is s=1

  real(8) :: haz(sim_ck,0:sim_cl) !(k,l) conditional hazard rate

contains
  subroutine setmodpar_june16 
  integer :: k,l,i,j
    par=new_modelpar(4,4) !par is initialized

    ! Parameters for data generating process:
    par%muw(1,:,1)   =(/ 3.72974d0, 4.09951d0, 4.38818d0, 4.90356d0 /)
    par%muw(2,:,1)   =(/ 4.09968d0, 4.25572d0, 4.42118d0, 4.59201d0 /)
    par%muw(3,:,1)   =(/ 4.42246d0, 4.37642d0, 4.49944d0, 4.64375d0 /)
    par%muw(4,:,1)   =(/ 4.71400d0, 4.61716d0, 4.78425d0, 4.97170d0 /)

    par%sigmaw(1,:,1)=(/ 1.39650d0, 0.83780d0, 0.77091d0, 0.86961d0 /)
    par%sigmaw(2,:,1)=(/ 0.55518d0, 0.25346d0, 0.25824d0, 0.27316d0 /)
    par%sigmaw(3,:,1)=(/ 0.14960d0, 0.11035d0, 0.12590d0, 0.14074d0 /)
    par%sigmaw(4,:,1)=(/ 0.24482d0, 0.19982d0, 0.20989d0, 0.22763d0 /)

    par%lam(1,:,1)    = (/ 0.030620d0, 0.026576d0, 0.020441d0, 0.011363d0 /)
    par%lam(2,:,1)    = (/ 0.026620d0, 0.031576d0, 0.022441d0, 0.010263d0 /)
    par%lam(3,:,1)    = (/ 0.020620d0, 0.021576d0, 0.036441d0, 0.030363d0 /)
    par%lam(4,:,1)    = (/ 0.019620d0, 0.022576d0, 0.030441d0, 0.037363d0 /)
    
    par%gam(1,:,1)   = (/ 0.07662d0, 0.04665d0, 0.13741d0, 0.73931d0 /)
    par%gam(2,:,1)   = (/ 0.06882d0, 0.06212d0, 0.18165d0, 0.68741d0 /)
    par%gam(3,:,1)   = (/ 0.04751d0, 0.02862d0, 0.14345d0, 0.78043d0 /)
    par%gam(4,:,1)   = (/ 0.03413d0, 0.01429d0, 0.11497d0, 0.83661d0 /)
    do k=1,par%k
        par%gam(k,:,1) = par%gam(k,:,1)/sum(par%gam(k,:,1))
    end do

    par%piw(:,1)      = (/ 0.8d0, 0.2d0 /)
    par%piw(:,2)      = (/ 0.7d0, 0.3d0 /)
    par%piw(:,3)      = (/ 0.5d0, 0.5d0 /)
    par%piw(:,4)      = (/ 0.3d0, 0.7d0 /)

    par%pif(1:sim_sdim,1)      = (/ 0.1d0, 0.9d0 /)
    par%pif(1:sim_sdim,2)      = (/ 0.2d0, 0.8d0 /)
    par%pif(1:sim_sdim,3)      = (/ 0.4d0, 0.6d0 /)
    par%pif(1:sim_sdim,4)      = (/ 0.6d0, 0.4d0 /)

		par%pif(0,0) =1.d0

    par%q       = (/ 0.00000d0, 0.10000d0, 0.50000d0, 0.30000d0, 0.10000d0 /)

!		par%p				= (/ 0.2d0, 0.4d0, 0.3d0, 0.1d0 /)

    par%psi(1,:,1)=(/ 0.01007d0, 0.00616d0, 0.00959d0, 0.00643d0 /)
    par%psi(2,:,1)=(/ 0.00787d0, 0.00761d0, 0.01088d0, 0.00640d0 /)
    par%psi(3,:,1)=(/ 0.01830d0, 0.01779d0, 0.01675d0, 0.00735d0 /)
    par%psi(4,:,1)=(/ 0.01306d0, 0.00905d0, 0.01491d0, 0.00724d0 /)

    par%del(1,:,1)=(/ 0.00635d0, 0.00742d0, 0.00855d0, 0.00686d0 /)
    par%del(2,:,1)=(/ 0.00774d0, 0.00668d0, 0.00712d0, 0.00711d0 /)
    par%del(3,:,1)=(/ 0.00117d0, 0.00152d0, 0.00062d0, 0.00032d0 /)
    par%del(4,:,1)=(/ 0.00447d0, 0.00595d0, 0.00266d0, 0.00114d0 /)
    
    par%rho=0.531d0

    call M_update(par)
    
!    call steadystate(par) !updates par%m to the steadystate distribution.

    par%m(1,:,1)=(/ .02d0, 0.07d0, 0.03d0, 0.02d0, 0.01d0 /)
    par%m(2,:,1)=(/ .02d0, 0.03d0, 0.09d0, 0.05d0, 0.02d0 /)
    par%m(3,:,1)=(/ .04d0, 0.03d0, 0.06d0, 0.11d0, 0.05d0 /)
    par%m(4,:,1)=(/ .04d0, 0.03d0, 0.07d0, 0.09d0, 0.12d0 /)

		par%m=par%m/sum(sum(par%m(:,:,1),dim=2))

		par%p=sum(par%m(:,:,1),dim=2)

		par%pz=0.d0
		do i=1,sim_zdim
			do k=1,sim_ck
				par%pz(i)=par%pz(i)+par%piw(i,k)*par%p(k)
			end do
		end do

		par%ps=0.d0
		do i=1,sim_sdim
			do l=1,sim_cl
				par%ps(i)=par%ps(i)+par%pif(i,l)*par%q(l)
			end do
		end do

		sprob=par%ps(1) !prob of s=1.
    zprob=par%pz(1) !prob of z=1

    truth=new_modelpar(sim_ck,sim_cl)
    truth=par
    
  end subroutine setmodpar_june16
  
  subroutine cleansim
    integer :: i
    do i=1,dim_I !clean up the simulated spells
       call remove_spell(work(i)%first)
       work(i)%first=>null()
       work(i)%last=>null()
    end do
    do i=1,dim_J
       firm(i)%first=>null()
       firm(i)%last=>null()
    end do
    call modelpar_deallocate(par)
  end subroutine cleansim

  subroutine simdat(seed)
    use tools
    use marsaglia
    integer, intent(in) :: seed
    integer(4) :: i,j !counters
    integer(4) :: kk,ll
    integer(4) :: widcdf(0:sim_ck,sim_zdim),fidcdf(0:sim_cl)
    real(8) :: ssmat(sim_ck,0:sim_cl) !ssmat contains the steady state match cdf of (k,l).
    real(8) :: inicdf(sim_ck,-1:sim_cl) !Contains k type conditional steady state match cdf. sscdf(.,0) is by def zero.
    real(8) :: destcdf(sim_ck,0:sim_cl,-1:sim_cl) !(k,l) conditional destination cdf.
    real(8) :: kzcdf(sim_ck,sim_zdim),lscdf(sim_cl,sim_sdim)

    dim_I=mydim_I
    dim_J=mydim_J
    allocate(work(dim_I),firm(0:dim_J))
    do i=1,80
       yrtoweek(i)=(i-1)*yr+1
    end do
    !set RNG
    call setseed(kisseed(2345678+2*seed,87656789+200*seed,5434567+100*seed,43212345+500*seed))
    !First, take care of preliminaries. Determine distributions and firm types.
    do kk=1,par%k
       inicdf(kk,-1)=0.d0
       do ll=0,par%l
          inicdf(kk,ll)=inicdf(kk,ll-1)+par%m(kk,ll,1)
          haz(kk,ll)=1.d0-par%Mbar(kk,ll,1) !mym(kk,ll,0,par%l)
          destcdf(kk,ll,-1)=0.0d0
          do j=0,par%l
             destcdf(kk,ll,j)=destcdf(kk,ll,j-1)+par%MM(kk,ll,j,1) !mym(kk,ll,j)
          end do
          destcdf(kk,ll,:)=destcdf(kk,ll,:)/haz(kk,ll)
       end do
       inicdf(kk,:)=inicdf(kk,:)/inicdf(kk,par%l) !Each row now contains a CDF.
    end do

		do i=1,sim_zdim
			do kk=1,par%k
				kzcdf(kk,i)=par%piw(i,kk)*sum(par%m(kk,:,1))/par%pz(i)
			end do
			do kk=2,par%k
				kzcdf(kk,i)=kzcdf(kk-1,i)+kzcdf(kk,i)
			end do
			write(6,'("This number should be close to one (k), ",f12.6,i5)') kzcdf(par%k,i),i
			kzcdf(:,i)=kzcdf(:,i)/kzcdf(par%k,i)
		end do

		do i=1,sim_sdim
			do ll=1,par%l
				lscdf(ll,i)=par%pif(i,ll)*par%q(ll)/par%ps(i)
			end do
			do ll=2,par%l
				lscdf(ll,i)=lscdf(ll-1,i)+lscdf(ll,i)
			end do
			write(6,'("This number should be close to one (l), ",f12.6,i5)') lscdf(par%l,i),i
			lscdf(:,i)=lscdf(:,i)/lscdf(par%l,i)
		end do

    fidcdf(0)=0
    forall(ll=1:sim_cl-1) fidcdf(ll)=int(sum(par%q(1:ll))*dble(dim_J)) 
    fidcdf(sim_cl)=dim_J
!!$    if(master) write(6,'(i5,i15,f15.5)') (ll,fidcdf(ll),par%q(ll),ll=1,sim_cl)
    
		!Set firm types
    firm(0)%l=0
    firm(0)%s=0
    ll=1
    do j=1,dim_J
    	if(j>fidcdf(ll)) ll=ll+1
    	firm(j)%l=ll
			if(j-fidcdf(ll-1)<int(par%q(ll)*dble(dim_J)*par%pif(1,ll))+1) then
				firm(j)%s=1
			else
				firm(j)%s=2
			end if
    end do

    !Now, for the panel simulation and assignment of types to workers (assumption is 2 types)
    kk=1
    do i=1,dim_I*zprob
       if(i .gt. kzcdf(kk,1)*dim_I*zprob+1) kk=kk+1
       work(i)%z=1
       work(i)%k=kk !worker i's type
       if(mod(i-1,numprocs)==myid) call spellsim(inidraw(work(i)%k)) !simulate the worker's spell history
    end do
    kk=1
    do i=dim_I*zprob+1,dim_I
       if(i .gt. dim_I*zprob+kzcdf(kk,2)*dim_I*(1.d0-zprob)+1) kk=kk+1
       work(i)%k=kk !worker i's type
       work(i)%z=2
       if(mod(i-1,numprocs)==myid) call spellsim(inidraw(work(i)%k)) !simulate the worker's spell history
    end do
  contains
    recursive subroutine spellsim(firmtype) !use worker index i from parent.
      integer, intent(in) :: firmtype
      integer :: mycount
      integer :: dur,start,durnc,startyr,endyr,duryr
      call newspell !creates a spell and makes this point to it.
      this%wid=i
      if(firmtype==0) then
         this%fid=0
      else
         this%fid=fiddraw(firmtype) !random draw of firm type from ss, then assign firm id from group of type ll firms.
         if(firm(this%fid)%l .ne. firmtype) stop'type assignment failed'
      end if
      if(associated(work(i)%last)) then
         start=1+work(i)%last%end !set current spell's start date equal to previous's end date
      else
         start=1 !hello
      end if
      durnc=1+int(log(1.0d0-kissuni())/log(1.0d0-haz(work(i)%k,firmtype)))
      dur=min(durnc,1+sim_T-start) !duration of spell
      this%end=start+dur-1
      this%strt=start

      startyr=int(real(this%strt)/real(yr))+1
      if(this%strt<yrtoweek(startyr)) startyr=startyr-1
      this%yr=startyr
      
      endyr=int(real(this%end)/real(yr))+1
      if(this%end<yrtoweek(endyr)) endyr=endyr-1

      duryr=endyr-startyr+1
      if(duryr<=0) stop'this will end badly.'
      allocate(this%wage(duryr))
      this%wage=0.d0
      if(firmtype>0) this%wage=wagedraw(duryr,work(i)%k,firm(this%fid)%l)
      call add_worker(work(i),this)
      call add_firm(firm(this%fid),this)
      if(this%end<sim_T) then
         ll=findndx(destcdf(work(i)%k,firm(this%fid)%l,:),kissuni())-1 !determine type of destination firm.
         call spellsim(ll) !do a new spell duration and wages.
      end if
    end subroutine spellsim

    function inidraw(myk) !return a random firm type realization from myk's steady state match distribution.
      integer(1), intent(in) :: myk
      integer :: inidraw
      inidraw=findndx(inicdf(myk,-1:par%l),kissuni())-1 !1st element in inicdf(k,:) is g_k1, 2nd element is g_k1+g_k2, etc...
    end function inidraw

    function fiddraw(myl)
      integer, intent(in) :: myl
      integer :: fiddraw
      if(myl==0) then
         fiddraw=0
      else
         fiddraw=kissuni()*(fidcdf(myl)-fidcdf(myl-1))+fidcdf(myl-1)+1
      end if
    end function fiddraw

    function wagedraw(mydur,myk,myl) !normally distributed random draws with variance par%sigmaw(myk,myl)*2 and mean par%muw(myk,myl)
      integer(1), intent(in) :: myk
      integer, intent(in) :: mydur,myl
      real(8), dimension(mydur) :: wagedraw
      real(8) :: prev_ar, new_ar, epsilon
      integer :: myi
      do myi=1,mydur
        if(myi==1) then
         prev_ar = kissgaus()*par%sigmaw(myk,myl,1)
         wagedraw(myi) = prev_ar+par%muw(myk,myl,1)
        else
         epsilon = kissgaus()*par%sigmaw(myk,myl,1)
         new_ar = par%rho*prev_ar+epsilon
         wagedraw(myi) = new_ar+par%muw(myk,myl,1)
         prev_ar = new_ar
        end if
      end do
    end function wagedraw
  end subroutine simdat

!!$  recursive function mym(k,l_out,l_in,sumto) result(myres) !transition probability of moving from l_out to l_in given k.
!!$    integer, intent(in) :: l_out,l_in,k
!!$    integer, optional, intent(in) :: sumto !this turns mym into a summation from l_in to sumto.
!!$    real(8) :: myres
!!$    if(l_in==0 .and. l_out==0) then
!!$       myres=0.0d0
!!$    else
!!$       myres=par%lam(l_out,1)*par%nu(l_in,1)*par%gam(k,l_in,1)/(par%gam(k,l_out,1)+par%gam(k,l_in,1))
!!$    end if
!!$    if(present(sumto)) then
!!$       if(l_in .lt. sumto) myres=myres + mym(k,l_out,l_in+1,sumto)
!!$    end if
!!$  end function mym

  subroutine simcheck
    integer :: i,j,jj,ii,ll,cdur(par%k,0:par%l)
    real(8) :: iniemp,myemp,calcemp,mydur(par%k,0:par%l),totdur(par%k,0:par%l)
!    write(6,'(" ")')
    !check emp in initial state
    
    do ii=1,2 !by z
       do ll=0,par%l
          calcemp=0.d0
          do i=1,par%k
             calcemp=calcemp+par%piw(ii,i)*par%p(i)*par%m(i,ll,1)
          end do
          j=0
          jj=0
          do i=1,dim_I
             if(associated(work(i)%first)) then
                if(work(i)%z==ii) jj=jj+1
                if(firm(work(i)%first%fid)%l==ll .and. work(i)%z==ii) j=j+1
             end if
          end do
          myemp=dble(j)/dble(jj)
          call mpi_reduce(myemp,iniemp,1,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
          if(master) write(6,'(3i5,2f20.10)') myid,ii,ll,calcemp,iniemp/dble(numprocs)
       end do
    end do

    cdur=0
    mydur=0.d0
    do i=1,dim_I
       call setdur(work(i)%first)
    end do
    mydur=mydur/dble(cdur)
    call mpi_reduce(mydur,totdur,(par%l+1)*par%k,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    totdur=totdur/dble(numprocs)
    if(master) write(6,'(2I5,2f15.5)') ((i,j,totdur(i,j),1.d0/haz(i,j),j=0,par%l),i=1,par%k)

  contains
    recursive subroutine setdur(mysp)
      type(spell), pointer :: mysp
      if(.not. associated(mysp)) return
      mydur(work(mysp%wid)%k,firm(mysp%fid)%l)=mydur(work(mysp%wid)%k,firm(mysp%fid)%l)+dur(mysp)
      cdur(work(mysp%wid)%k,firm(mysp%fid)%l)=cdur(work(mysp%wid)%k,firm(mysp%fid)%l)+1
      if(associated(mysp%next)) call setdur(mysp%next)
    end subroutine setdur
  end subroutine simcheck
end module datbysim
