module modelpar_class !contains model parameters and objects and operations on them.
  implicit none
  integer, parameter :: hetdim=8    !wage parameter x het dim 
  integer, parameter :: inihetdim=8 !m dist x het dim 
  integer, parameter :: mobhetdim=8 !mob parameter x het dim 
  integer, parameter :: sdhetdim=12 !sd parameter x het dim 
  integer, parameter :: zhetdim=6   !z dim.
  integer, parameter :: shetdim=36   !s dim (number of industries)

  type pardim
     integer :: k
     integer :: l
     integer :: z
     integer :: s
     integer :: het
     integer :: sig
     integer :: mob
     integer :: ini
  end type pardim
  
  type modelpar
     integer :: k,l
     type(pardim) :: dim
     real(8),allocatable :: muw(:,:,:),sigmaw(:,:,:) !worker and firm type specific (unobservables) wage mean and std dev.
     real(8),allocatable :: gam(:,:,:)  !match surplus means by firm and worker unobservables: worker type specific ranks.
     real(8),allocatable :: lam(:,:,:)  !(k,l) specific offer arrival rate.
     real(8),allocatable :: m(:,:,:)    !initial distribution m(k,l|x) (x conditional)
     real(8),allocatable :: q(:)        !distribution of firm types
     real(8),allocatable :: pk(:)       !distribution of worker types
     real(8),allocatable :: del(:,:,:)  !(k,l,x) conditional layoff rate.
     real(8),allocatable :: psi(:,:,:)  !probability that a (k,x) individual finds a l' job.
     real(8),allocatable :: MM(:,:,:,:) !M(k,l,ll,x) prob that type k worker moves from l to ll given x
     real(8),allocatable :: Mbar(:,:,:) !\bar{M}(k,l,x) prob that type k worker stays with l given x
     real(8),allocatable :: piw(:,:)    !pi_w(z|k)
     real(8),allocatable :: pif(:,:)	!pi_f(s|l). s is the industry x sector indicator. by default s=0 is a non-category for l=0.
     !pif needs to handle l=0 matches and so dim of pif runs pif(0:shetdim,0:myl), where pif(0,0)=1.
     real(8),allocatable :: pz(:) !marginal distribution of z
     real(8),allocatable :: ps(:) !marginal distribution of s
     real(8) :: rho !AR(1) parameter
  end type modelpar
contains
  function capM(k,l,ll,x,mypar) result(res) !transition probability of moving from l to ll given k.
    integer, intent(in) :: l,ll,k,x
    type(modelpar), intent(in) :: mypar
    real(8) :: res
    if(l==0 .and. ll==0) then !M_{k00}(x)=0 by definition
       res=0.0d0
    else if(l==0) then !Move from unemployment to emp.
       res=mypar%psi(k,ll,x)
    else if (ll==0) then !layoff
       res=mypar%del(k,l,x)
    else
       if(mypar%gam(k,l,x)+mypar%gam(k,ll,x)>0.d0) then
          res=mypar%lam(k,ll,x)*mypar%gam(k,ll,x)/(mypar%gam(k,l,x)+mypar%gam(k,ll,x))
       else
          res=0.5d0*mypar%lam(k,ll,x)
       end if
    end if
  end function capM

  subroutine M_update(par)
    type(modelpar), intent(inout) :: par
    integer :: l,ll,k,x
    do x=1,mobhetdim
       do k=1,par%k
          do l=0,par%l
             do ll=0,par%l
                par%MM(k,l,ll,x)=capM(k,l,ll,x,par)
             end do
             par%mbar(k,l,x)=1.d0-sum(par%MM(k,l,:,x))
             !if mbar<=0, set it to a small numner and rescale par%MM too
             if(par%mbar(k,l,x)<=0.d0) then
                par%mbar(k,l,x)=0.000001d0
                par%MM(k,l,:,x)=par%MM(k,l,:,x)/sum(par%MM(k,l,:,x))*0.999999d0
             end if
          end do
       end do
    end do
  end subroutine M_update

  subroutine readmodelpar(ch,mypar,myk,myl)
    type(modelpar), intent(inout) :: mypar
    integer, intent(in) :: myk,myl,ch
    integer :: kmap(myk),lmap(0:myl)
    integer :: ii,jj,tt,kk
    read(ch,'(*(i5))') (kmap(ii),ii=1,myk)
    read(ch,'(*(i5))') (lmap(ii),ii=0,myl)    
    read(ch,'(i4)') mypar%k, mypar%l
    do tt=1,hetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%muw(ii,jj,tt), jj=1,myl)
       end do
    end do
    do tt=1,sdhetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%sigmaw(ii,jj,tt), jj=1,myl)
       end do
    end do
    do tt=1,mobhetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%gam(ii,jj,tt), jj=1,myl)
       end do
    end do
    do tt=1,mobhetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%lam(ii,jj,tt), jj=1,myl)
       end do
    end do
    do tt=1,mobhetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%psi(ii,jj,tt), jj=1,myl)
       end do
    end do
    do tt=1,mobhetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%del(ii,jj,tt), jj=1,myl)
       end do
    end do
    do tt=1,zhetdim
       read(ch,'(*(f15.8))') (mypar%piw(tt,ii),ii=1,myk)
    end do
    do tt=1,shetdim
       read(ch,'(*(f15.8))') (mypar%pif(tt,jj),jj=1,myl)
    end do

    do tt=1,inihetdim
       do ii=1,myk
          read(ch,'(*(f15.8))') (mypar%m(ii,jj,tt), jj=0,myl)
       end do
    end do
    read(ch,'(f15.8)') (mypar%q(jj),jj=1,myl)
    read(ch,'(f15.8)') (mypar%pk(ii),ii=1,myk)
    read(ch,'(f15.8)') (mypar%pz(jj),jj=1,zhetdim)
    read(ch,'(f15.8)') (mypar%ps(jj),jj=1,shetdim)
    read(ch,'(f15.8)') mypar%rho
    do tt=1,mobhetdim
       do kk=1,mypar%k
          do ii=0,mypar%l
             read(ch,'(*(f15.8))') (mypar%MM(kk,ii,jj,tt), jj=0,mypar%l)
          end do
       end do
    end do
    do tt=1,mobhetdim
       do kk=1,mypar%k
          read(ch,'(*(f15.8))') (mypar%Mbar(kk,jj,tt), jj=0,mypar%l)
       end do
    end do
  end subroutine readmodelpar
  
  function new_modelpar(myk,myl) result(res)
    integer, intent(in) :: myk,myl
    type(modelpar) :: res
    allocate(res%muw(myk,myl,hetdim),res%sigmaw(myk,myl,sdhetdim),res%gam(myk,myl,mobhetdim),res%lam(myk,myl,mobhetdim),&
         &res%m(myk,0:myl,inihetdim),res%q(0:myl),res%pk(myk),res%del(myk,myl,mobhetdim),&
         &res%psi(myk,myl,mobhetdim),res%MM(myk,0:myl,0:myl,mobhetdim),res%Mbar(myk,0:myl,mobhetdim),&
         &res%piw(zhetdim,myk),res%pif(0:shetdim,0:myl),res%pz(zhetdim),res%ps(shetdim))
    res%muw   =0.d0
    res%sigmaw=0.d0
    res%gam   =0.d0
    res%lam   =0.d0
    res%m     =0.d0
    res%piw   =0.d0
    res%pif   =0.d0
    res%q     =0.d0
    res%pk    =0.d0
    res%del   =0.d0
    res%psi   =0.d0
    res%M     =0.d0
    res%Mbar  =0.d0
    res%q(0)  =1.d0 !unemp group. Not updated. By default one.
    res%k     =myk
    res%l     =myl
    res%rho   =0.d0
    res%pz    =0.d0
    res%ps    =0.d0
  end function new_modelpar

  function new_modelpar_new(mydim) result(res) !allocate modelpar according to mydim dimensions.
    type(pardim), intent(in) :: mydim
    type(modelpar) :: res
    res%dim=mydim
    allocate(res%muw(mydim%k,mydim%l,mydim%het),res%sigmaw(mydim%k,mydim%l,mydim%sig),res%gam(mydim%k,mydim%l,mydim%mob),&
         &res%lam(mydim%k,mydim%l,mydim%mob),res%m(mydim%k,0:mydim%l,mydim%ini),res%q(0:mydim%l),res%pk(mydim%k),&
         &res%del(mydim%k,mydim%l,mydim%mob),res%psi(mydim%k,mydim%l,mydim%mob),res%MM(mydim%k,0:mydim%l,0:mydim%l,mydim%mob),&
         &res%Mbar(mydim%k,0:mydim%l,mydim%mob),res%piw(mydim%z,mydim%k),res%pif(0:mydim%s,0:mydim%l),&
         &res%pz(mydim%z),res%ps(mydim%s))
    res%muw   =0.d0
    res%sigmaw=0.d0
    res%gam   =0.d0
    res%lam   =0.d0
    res%m     =0.d0
    res%piw   =0.d0
    res%pif   =0.d0
    res%q     =0.d0
    res%pk    =0.d0
    res%del   =0.d0
    res%psi   =0.d0
    res%M     =0.d0
    res%Mbar  =0.d0
    res%q(0)  =1.d0 !unemp group. Not updated. By default one.
    res%k     =mydim%k
    res%l     =mydim%l
    res%rho   =0.d0
    res%pz    =0.d0
    res%ps    =0.d0
  end function new_modelpar_new
  
  subroutine modelpar_deallocate(mypar)
    type(modelpar), intent(inout) :: mypar
    if(allocated(mypar%muw)) then
       deallocate(mypar%muw,mypar%sigmaw,mypar%gam,mypar%lam,mypar%m,mypar%pk,mypar%q,mypar%del,&
            &mypar%psi,mypar%MM,mypar%Mbar,mypar%piw,mypar%pif,mypar%pz,mypar%ps)
    end if
  end subroutine modelpar_deallocate
  
  function modpar_dist(in1,in2)
    type(modelpar), intent(in) :: in1,in2
    real(8) :: modpar_dist
    if(size(in1%muw) .ne. size(in2%muw)) then
       write(6,'("Fatal(modpar_dist): in1 and in2 do not have same dimensions.")')
       stop
    end if
    modpar_dist=sum(sum((in1%muw-in2%muw)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%sigmaw-in2%sigmaw)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%gam-in2%gam)**2,dim=1))
    modpar_dist=modpar_dist+sum((in1%lam-in2%lam)**2)
    modpar_dist=modpar_dist+(in1%rho-in2%rho)**2
    modpar_dist=modpar_dist+sum(sum((in1%m-in2%m)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%piw-in2%piw)**2,dim=1))
    modpar_dist=modpar_dist+sum(sum((in1%pif-in2%pif)**2,dim=1))
  end function modpar_dist

  function modparadd(mp1,mp2) result(res)
    type(modelpar), intent(in) :: mp1,mp2
    type(modelpar) :: res
    res=new_modelpar(mp1%k,mp1%l)
    res%muw   =mp1%muw   +mp2%muw
    res%sigmaw=mp1%sigmaw+mp2%sigmaw
    res%gam   =mp1%gam   +mp2%gam
    res%lam   =mp1%lam   +mp2%lam
    res%rho   =mp1%rho   +mp2%rho
  end function modparadd

  function modpardiv(mp1,xx) result(res)
    type(modelpar), intent(in) :: mp1
    real(8),intent(in) :: xx
    type(modelpar) :: res
    res=new_modelpar(mp1%k,mp1%l)
    res%muw   =mp1%muw/xx
    res%sigmaw=mp1%sigmaw/xx
    res%gam   =mp1%gam/xx
    res%lam   =mp1%lam/xx
    res%rho   =mp1%rho/xx
  end function modpardiv
end module modelpar_class
